/*
 * Copyright (c) 2009, Giampaolo Rodola'.
 * Copyright (c) 2015, Ryo ONODERA.
 * All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

PyObject *psutil_net_connections(PyObject *, PyObject *);
PyObject *psutil_proc_net_connections(PyObject *, PyObject *);
